/*
** Instrument.c - Standard betascan device startup
** rzookol & geit & deadwood
** based on Svend Daugaard Pedersen work
** (c) 2009 - 2019 Michal Zukowski
** (C) 2020 Harry Sintonen
*/

/// Includes
#define USE_INLINE_STDARG
#include <dos/dos.h>
#include <dos/dostags.h>
#include <exec/devices.h>
#include <exec/errors.h>
#include <exec/execbase.h>
#include <exec/initializers.h>
#include <exec/io.h>
#include <exec/lists.h>
#include <exec/memory.h>
#include <exec/nodes.h>
#include <exec/ports.h>
#include <exec/resident.h>
#include <exec/tasks.h>
#include <exec/types.h>
#include <libraries/dos.h>
#include <proto/dos.h>
#include <proto/exec.h>
#include <proto/intuition.h>
#include <clib/alib_protos.h>
#if !defined(__AROS__)
#include <proto/misc.h>
#include <proto/query.h>
#include <clib/debug_protos.h>
#else
#include <aros/symbolsets.h>
#endif

#include <stdio.h>
#include <string.h>
#include <instrument.h>


#include <DriverIo.h>
#include <devices/scanner.h>
#include <scanner_protos.h>

//#define DEBUG

///
/// Defines
#ifdef DEBUG
    #define DBUG(x)   x
#else
    #define DBUG(x)
#endif

#define BAR1(x) #x"."
#define BAR(x) #x
#define VERS(a, b) BAR1(a)BAR(b)

#include "epsonusb_intern.h"

#if defined(__AROS__)
#include LC_LIBDEFS_FILE
#endif

#define BAR1(x) #x"."
#define BAR(x) #x
#define VERS(a, b) BAR1(a)BAR(b)

#ifndef __AMIGADATE__
#define __AMIGADATE__ __DATE__
#endif

///

#ifdef __MORPHOS__
const ULONG __abox__ = 1;
const ULONG __amigappc__ = 1;

/* TLS implementation related */
static VOID tlsdestructor(APTR value);
static ULONG tlsindex;
#endif

char* versionString = "$VER: " DEVNAME " " VERS(V_MAJOR,V_MINOR) " (" __AMIGADATE__ ")";
char  DevName[] = DEVNAME;
char  DevIdString[] = DEVNAME;


/***************************************************************************/

struct ScannerIOInternal
{
  struct InstrumentIO   IOScan;
  struct ScannerOptions option;
};

struct UnitId
{
  struct Node     ui_Node;
  struct Process *ui_Prc;
  struct Unit     ui_Unit;
};

/***************************************************************************/

#ifdef __MORPHOS__

/*
** prototypes
*/
int                    DevOpen   ( void );
BPTR                   DevClose  ( void );
struct InstrumentBase *DevInit   ( struct InstrumentBase *libbase, BPTR seglist,struct ExecBase *sysbase);
BPTR                   DevExpunge( void );
void                   DevBeginIO( void );
int                    DevAbortIO( void );

/***************************************************************************/

/* /// Device Resident structure and other stuff
**
*/
static int ble(void)
{
    return -1;
}

static int DevReserved(void)
{
    return 0;
}
 /*
struct {

   INITBYTE(ln_type, NT_DEVICE)
   INITLONG(ln_Name, DevName),
   INITBYTE(lib_Flags,LIBF_SUMUSED|LIBF_CHANGED),
   INITWORD(lib_Version,VERSION )
   INITWORD(lib_Revision, REVISIOND)
   INITLONG(lib_IdString, DevIdString)
}
  */

static const ULONG FuncTable[] = {
    FUNCARRAY_32BIT_NATIVE,
    (ULONG)DevOpen,
    (ULONG)DevClose,
    (ULONG)DevExpunge,
    (ULONG)DevReserved,
    (ULONG)DevBeginIO,
    (ULONG)DevAbortIO,
    ~0
};

const struct {
    ULONG  LibSize;
    const void  *FuncTable;
    const void  *DataTable;
    void (*InitFunc)(void);
} init = {
    sizeof(struct InstrumentBase),
    FuncTable,
    NULL,
    (void (*)())DevInit
};

static const struct TagItem	QueryTags[] =
{
	{ QUERYINFOATTR_NAME, (ULONG) DevName },
	{ QUERYINFOATTR_IDSTRING, (ULONG) DevIdString },
	{ QUERYINFOATTR_DESCRIPTION, (ULONG) "SANE scanner driver for " BACKEND_NAME " compatible scanners." },
	{ QUERYINFOATTR_DATE, (ULONG) __AMIGADATE__ },
	{ QUERYINFOATTR_VERSION, (ULONG) V_MAJOR },
	{ QUERYINFOATTR_REVISION, V_MINOR },
    { QUERYINFOATTR_CODETYPE, MACHINE_PPC },
	{ QUERYINFOATTR_SUBTYPE, QUERYSUBTYPE_DEVICE },
	{ QUERYINFOATTR_CLASS, QUERYCLASS_SCANNER },
	{ QUERYINFOATTR_SUBCLASS, QUERYSUBCLASS_SCANNER_SANE },
    { TAG_DONE, 0 }
};


const struct Resident ROMTag = {
    RTC_MATCHWORD,
    (APTR)&ROMTag,
    (APTR)(&ROMTag + 1),
    RTF_PPC | RTF_AUTOINIT,
    V_MAJOR,
    NT_DEVICE,
    0,
    DEVNAME,
	DEVNAME " " VERS(V_MAJOR,V_MINOR) " (" __AMIGADATE__ ")\r\n",
	(APTR)&init
    /* New Fields */
	, V_MINOR,
    (struct TagItem*) &QueryTags
};
/* \\\ */


/* old stuff, check if needed */

struct ExecBase *SysBase;
#ifdef __MORPHOS__
struct Library *PsdBase = NULL;
#elif __AROS__
struct Library *ps = NULL;
#endif
//struct DOSBase *DOSBase;
#endif

extern int  stackSize;
extern char DevIdString[];

// Unit stuff

/// Unit_QuickIO()

static BOOL Unit_QuickIO( struct InstrumentIO *io)
{
	switch( io->io_Command ) {
	    case FCMD_OPEN:
	    case FCMD_CLOSE:
	    case SCANCMD_CONTROL:
	    case SCANCMD_READOPTION:
	    case CMD_STOP:
	    case CMD_START:
	    case CMD_READ:
			io->io_Flags &= ~IOF_QUICK;
			return FALSE;
	    default:
			io->io_Error = IOERR_NOCMD;
			return TRUE;
	}
}
///
/// Unit_Task()

static void Unit_Task( void )
{
struct Unit *Unit;
char devName[1024];
int devUnit;
int done = 0;
struct ScannerIOInternal* ior;
struct InstrumentIO* io;
struct Process *pr;
struct StartUpMsg *msg;
OpenRetVals ret;

	pr = (struct Process *) SysBase->ThisTask;

	WaitPort( &pr->pr_MsgPort );  // wait for startup message

	msg = ( struct StartUpMsg *) GetMsg( &pr->pr_MsgPort );

	Unit    = msg->unit;         /* obtain unit structure from message */
	devUnit = msg->devUnit;      /* obtain unit number from message */

	strcpy( devName, msg->devName ); /* copy name, as its only valid until reply startup message */

/* setup unit port and activate signal mode */

	Unit->unit_MsgPort.mp_SigBit = AllocSignal( -1 ); /* this should never fail as this task is band new */
	Unit->unit_MsgPort.mp_Flags  = PA_SIGNAL;

	ReplyMsg( &msg->msg ); /* reply startup message */

/* Get opening message (IO with io_Command = FCMD_OPEN) */

	WaitPort( &Unit->unit_MsgPort );

	ior = (struct ScannerIOInternal *) GetMsg( &Unit->unit_MsgPort );

	if( ior->IOScan.io_Message.mn_Length != sizeof( struct ScannerIOInternal ) ) {
		Unit->unit_OpenCnt = 0;
		ior->IOScan.io_Error = SCAN_OPNERR_FATAL;

		DBUG(KPrintF("ioReq message size %d doesn't match %d (sizeof(struct ScannerIOInternal)), replying msg %p\n",ior->IOScan.io_Message.mn_Length, sizeof(struct ScannerIOInternal), ior);)

		ReplyMsg( (struct Message *) ior );
	}

/* setup our unit */

	ior->IOScan.io_Unit = &Unit;

/* setup hardware */
	DBUG(KPrintF("open driver IO\n");)
	if(( ret = openDriverIo( devName, devUnit, &Unit->unit_MsgPort )) != Open_Failure) 
	{	
		if (ret == Open_USB)
		{
			// opening poseidon
			#ifdef __MORPHOS__
			if(!PsdBase  && !(PsdBase = OpenLibrary("poseidon.library", 1)))
			#elif  __AROS__
			if(!ps && !(ps= OpenLibrary("poseidon.library", 1)))
			#endif
			{
				ior->IOScan.io_Error = SCAN_OPNERR_DEVICE;
			}
			else
			{
				openScanner( &ior->option, &ior->IOScan.io_Error );

				DBUG(KPrintF("openScanner returned\n");)

				if( ior->IOScan.io_Error ) {
					closeDriverIo();
				}
			}
		}
	} 
	else 
	{
		ior->IOScan.io_Error = SCAN_OPNERR_DEVICE;
	}

/* if hardware is up and running, launch task main loop */

	if( ior->IOScan.io_Error == SCAN_STATUS_OK ) {
		Unit->unit_OpenCnt = 1;
            //      Unit->unit_OpenCnt++;

		DBUG(KPrintF("scan status SCAN_STATUS_OK, replying msg %p\n",ior);)

		ReplyMsg((struct Message *) ior );

		while( !done  ) {
			while( ( io = (struct InstrumentIO*) waitForIOReq() ) ) {

				DBUG(KPrintF("%s: received io_Command %d\n",__FUNCTION__,io->io_Command);)

				switch(io->io_Command) {
					case FCMD_OPEN:
						io->io_Error = IOERR_UNITBUSY; /* each unit can only be opened once, so return error */
						break;
					case FCMD_CLOSE:
					    //	  Unit->unit_OpenCnt--;
                                                Unit->unit_OpenCnt = 0L;
					    //	  if( Unit->unit_OpenCnt == 0 )
                                                {
                                                        done = 1;
						}
						break;
					case SCANCMD_READOPTION:
						readOption((struct OptionDescriptor *) io->io_Data, &io->io_Error );
						break;
					case SCANCMD_CONTROL:
						controlOption( (struct OptionValue *) io->io_Data, &io->io_Error );
						break;
					case CMD_STOP:
						stopScanning();
						io->io_Error = 0;
						break;
					case CMD_START:
						startScanning( (struct ScanInformation*) io->io_Data, &io->io_Error );
						break;
					case CMD_READ:
						io->io_Actual = io->io_Length;
						readScanLine( (struct ScanLine*) io->io_Data, &io->io_Error );
						break;
					default:
						io->io_Error = IOERR_NOCMD;
						break;
				}

				ReplyMsg( (struct Message*) io );
					if (done)
						break;
			}
            DBUG(KPrintF("ioreq end !!!\n"));
		}
                DBUG(KPrintF("TASK: closeScanner start \n"));
		closeScanner();
                DBUG(KPrintF("TASK: closeScanner end \n"));
		closeDriverIo();
                DBUG(KPrintF("TASK: closedriver\n"));
		WaitPort( &pr->pr_MsgPort );  // wait for shutdown message
                DBUG(KPrintF("TASK: waitport\n"));
		msg = ( struct StartUpMsg *) GetMsg( &pr->pr_MsgPort );
                DBUG(KPrintF("TASK: getmsg\n"));
		ReplyMsg( &msg->msg ); /* reply shutdown message */
                DBUG(KPrintF("TASK: ReplyMsg\n"));
	} else { /* on error during FCMD_OPEN we simply reply and kill task */
		Unit->unit_OpenCnt = 0;
		ReplyMsg( (struct Message *) ior );
	}
}
///
/// Unit_Alloc()

static struct UnitId *Unit_Alloc( struct InstrumentBase *libbase, UBYTE *name, int num )
{
    struct UnitId *result;
    char PrcName[256];
	char *p;
    struct UnitId *id = NULL;

	p =  stpcpy( PrcName, DevIdString );
	p =  stpcpy( p, "." );
	p =  stpcpy( p, (char *) name );
	p =  stpcpy( p, ".");
	sprintf(p, "%d", num);  // adding unit number to process name


	ObtainSemaphore( &libbase->vb_Semaphore );
	if( !(result = (struct UnitId *) FindName ( &libbase->vb_UnitList, PrcName ) ) ) {
		struct StartUpMsg msg;
		if( (id = (struct UnitId *)AllocVec( sizeof( struct UnitId ) + strlen( PrcName ) + 1, MEMF_CLEAR ) ) ) {
			stpcpy( (char *) (id + 1 ), PrcName );
			id->ui_Node.ln_Name = (char *) ( id + 1 );

			if( ( msg.msg.mn_ReplyPort = CreateMsgPort() ) ) {
#if !defined(__AROS__)
				if( ( id->ui_Prc = CreateNewProcTags( NP_Entry, (IPTR) &Unit_Task,
				                                      NP_PPCStackSize, stackSize,
				                                      NP_CodeType, CODETYPE_PPC,
				                                      NP_Name,(IPTR) (id+1),
				                                      NP_Priority,4,
				                                      TAG_DONE ) ) ) {
#else
				if( ( id->ui_Prc = CreateNewProcTags( NP_Entry, &Unit_Task,
				                                      NP_StackSize, stackSize,
				                                      NP_Name,(char *)(id+1),
				                                      NP_Priority,4,
				                                      TAG_DONE ) ) ) {
#endif
				    id->ui_Unit.unit_MsgPort.mp_Node.ln_Type = NT_MSGPORT;
					id->ui_Unit.unit_MsgPort.mp_Node.ln_Name = (char *) ( id + 1 );
				    id->ui_Unit.unit_MsgPort.mp_Flags = PA_IGNORE;
					id->ui_Unit.unit_MsgPort.mp_SigTask = (struct Task *) id->ui_Prc;
					NewList(&id->ui_Unit.unit_MsgPort.mp_MsgList);

				    id->ui_Unit.unit_OpenCnt = 0;
				    /* Send the startup message */
					msg.msg.mn_Length       = sizeof(struct StartUpMsg);
				    msg.msg.mn_Node.ln_Type = NT_MESSAGE;
					msg.unit                = &id->ui_Unit;
					msg.devName             = name;
					msg.devUnit             = num;

					PutMsg( &id->ui_Prc->pr_MsgPort, &msg.msg );         /* put startup message, so task can init itself */
					WaitPort( msg.msg.mn_ReplyPort );                    /* wait for task reply */
					GetMsg( msg.msg.mn_ReplyPort );						 /* get task reply (no need to keep message) */

					if( msg.unit ) {
						AddTail( &libbase->vb_UnitList, &id->ui_Node );  /* unit is still set, so task init succeeded */
						result = id;                                     /* set result */
					} else {                                             /* something went wrong, so send shut down message */
				      msg.msg.mn_Node.ln_Type = NT_MESSAGE;

					  PutMsg( &id->ui_Prc->pr_MsgPort, &msg.msg );
					  WaitPort( msg.msg.mn_ReplyPort );
					  GetMsg( msg.msg.mn_ReplyPort );                    /* get task reply (no need to keep message) */
				      FreeVec(id);
				      id = NULL;
				    }
				}
				DeleteMsgPort( msg.msg.mn_ReplyPort );                   /* kill port */
			}
		}
	}
	if( !result ) {
		if( id ) {
		  FreeVec( id );
		}
    }
	ReleaseSemaphore( &libbase->vb_Semaphore );
	return( result );
}
///
/// Unit_Free()

static void Unit_Free( struct InstrumentBase *libbase, struct UnitId *id )
{
	struct Message msg;
    
	DBUG(KPrintF("Unit_Free: start\n"));
	if( ( msg.mn_ReplyPort = CreateMsgPort() ) ) {

		msg.mn_Length       = sizeof( struct Message );
		msg.mn_Node.ln_Type = NT_MESSAGE;

		ObtainSemaphore( &libbase->vb_Semaphore );
                DBUG(KPrintF("Unit_Free: ObtainSemaphore\n"));
		PutMsg( &id->ui_Prc->pr_MsgPort, &msg );

		WaitPort( msg.mn_ReplyPort );
                DBUG(KPrintF("Unit_Free: WaitPort\n"));
		GetMsg( msg.mn_ReplyPort );

		DeleteMsgPort( msg.mn_ReplyPort );

		Remove( &id->ui_Node );

		FreeVec( id );

		ReleaseSemaphore( &libbase->vb_Semaphore );
	}
	DBUG(KPrintF("Unit_Free: end\n"));
}
///

// Device functions

/// DevInit()

#if !defined(__AROS__)
struct InstrumentBase *DevInit( struct InstrumentBase *libbase,		       
				BPTR seglist,
				struct ExecBase *sysbase)
{

	SysBase = (*((struct ExecBase **) 4 ) );
#else
static int GM_UNIQUENAME(Init)(LIBBASETYPEPTR libbase)
{
#endif

        DBUG(KPrintF("Devinit start\n"));

	if( SysBase->LibNode.lib_Version < 37 ) {
#if !defined(__AROS__)        
		return( NULL ); /* can only run under 2.04 or greater */
#else
        return FALSE;
#endif
	} else {
		InitSemaphore( &libbase->vb_Semaphore );
		NewList( &libbase->vb_UnitList );

#ifdef __MORPHOS__
		/* If TLS functionality is available use it for per-task data. */

		if (LIB_MINVER(&SysBase->LibNode, 51, 46))
			tlsindex = TLSAlloc(TLSTAG_DESTRUCTOR, (IPTR) tlsdestructor, TAG_DONE);
		else
			tlsindex = TLS_INVALID_INDEX;
#endif

#if !defined(__AROS__)          
                libbase->vb_SegList =seglist;
		return( libbase );
	}
}
#else
        return TRUE;
    }
}
#endif

///
/// DevOpen()

#ifdef __MORPHOS__
int DevOpen( void )
{
struct InstrumentBase *libbase = (APTR) REG_A6;
long UnitNum = REG_D0;
struct InstrumentIO *ior = (struct InstrumentIO *) REG_A1;
//long flags  = REG_D1;
#else
static int GM_UNIQUENAME(Open)
(
    LIBBASETYPEPTR libbase,
    struct IORequest *ioreq,
    ULONG UnitNum,
    ULONG flags
)
{
struct InstrumentIO *ior = (struct InstrumentIO*)ioreq;
#endif
struct UnitId *id;
UBYTE* DevName;
int retcode;
  
        DBUG(KPrintF("Devopen: start\n"));
	DevName = (UBYTE *)( ior->io_Data );

	retcode = IOERR_OPENFAIL;

	libbase->vb_Library.lib_OpenCnt++;  /* increase open count */

    if (!DOSBase) {
		 DOSBase = (APTR) OpenLibrary("dos.library", 0 );
	 }
	 if( DOSBase) {

		if( (id = Unit_Alloc( libbase, DevName, UnitNum ) ) ) {
			struct MsgPort *TimerPort = NULL;
			struct timerequest *TimerReq = NULL;
			ULONG timesig=0;
			int pendingTimer=0;
			struct MsgPort *OldReply;
	                DBUG(KPrintF("DevOpen: unit allloc starts\n"));
			OldReply = ior->io_Message.mn_ReplyPort;

			if ((TimerPort = CreateMsgPort())) {
				if ((TimerReq = (struct timerequest *)CreateIORequest(TimerPort, sizeof(struct timerequest)))) {
					if (0 == OpenDevice(TIMERNAME, UNIT_VBLANK, (struct IORequest *)TimerReq, 0)) {
						timesig = 1UL << TimerPort->mp_SigBit;
						TimerReq->tr_node.io_Command = TR_ADDREQUEST;
					} else {
						DBUG(KPrintF("Unable to open %s!\n", (ULONG)TIMERNAME);)
					}
				} else {
					DBUG(KPrintF("Unable to create IORequest!\n");)
				}
			}

			if( ( ior->io_Message.mn_ReplyPort = CreatePort( NULL, 0 ) ) ) {
				ULONG sigrcvd;
				ULONG replysig =  1UL << ior->io_Message.mn_ReplyPort->mp_SigBit;
	                        DBUG(KPrintF("DevOpen: create port starts\n"));
				ior->io_Error = 0;
				ior->io_Command = FCMD_OPEN;
				PutMsg( &id->ui_Unit.unit_MsgPort, &ior->io_Message );
			        DBUG(KPrintF("DevOpen: putmsg\n"));
				{ //WaitPort(ior->io_Message.mn_ReplyPort); This should be fixed without Delay() crap
				struct Message *msg=NULL;

				if(TimerReq) {
					TimerReq->tr_time.tv_secs = 30;	// timeout timer
					TimerReq->tr_time.tv_micro = 0;
	        
					SendIO((struct IORequest *)TimerReq);
					pendingTimer++;
				}

				sigrcvd = Wait(timesig | replysig);

				// we just care about what we get first, either timeout or msg  ...

				if(sigrcvd & timesig) {
					// got timer sig, timeout ...

					if(pendingTimer) {
						WaitIO((struct IORequest *)TimerReq);
						pendingTimer = 0;
		                             	DBUG(KPrintF("DevOpen: got timeout\n"));
					}
				}			

				if(sigrcvd & replysig) {
					msg = GetMsg( ior->io_Message.mn_ReplyPort );
	                             	DBUG(KPrintF("DevOpen: got replymsg 0x%p Command %d\n",msg,((struct IORequest *)msg)->io_Command));
				}


				if( msg == NULL ) {
	          // Hanging - send abort signal
	          		       DBUG(KPrintF("DevOpen: msg == null \n"));
				    //	  Signal( &id->ui_Prc->pr_Task, SIGBREAKF_CTRL_C );
					WaitPort( ior->io_Message.mn_ReplyPort );
				}
				}
	                        DBUG(KPrintF("DevOpen: delete ports starts\n"));
				DeletePort( ior->io_Message.mn_ReplyPort );
				if( ior->io_Error ) { 
                                        Remove( &id->ui_Node );
					FreeVec( id );
				    //	Unit_Free( libbase, id );  /* on error we free new unit again, as we failed */
					retcode = ior->io_Error;
				} else {
					retcode = 0;
				}
			}

			if (pendingTimer) {
				AbortIO((struct IORequest *)TimerReq);
				WaitIO((struct IORequest *)TimerReq);
			}

			if (timesig) CloseDevice((struct IORequest *)TimerReq);
			if (TimerReq) DeleteIORequest((struct IORequest *)TimerReq);
			if (TimerPort) DeleteMsgPort(TimerPort);

			ior->io_Message.mn_ReplyPort = OldReply;
		}
		DBUG(KPrintF("DevOpen: end retcode=%d\n",retcode));
	}
/* on error, we didn't open a unit, so we need to decrease opencnt again */
	if( retcode ) {
		libbase->vb_Library.lib_OpenCnt--;
	}
#if !defined(__AROS__)
	return( retcode );
#else
    if (retcode == 0) return TRUE; else return FALSE;
#endif
}

///
/// DevClose()

#if !defined(__AROS__)
BPTR DevClose( void)
{
struct InstrumentBase *libbase = (APTR) REG_A6;
struct InstrumentIO *ior = (APTR) REG_A1;
#else
static int GM_UNIQUENAME(Close)
(
    LIBBASETYPEPTR libbase,
    struct IORequest *ioreq
)
{
struct InstrumentIO *ior = (struct InstrumentIO*)ioreq;
#endif
struct UnitId *id;
 
       DBUG(KPrintF("DevClose: start\n"));
/* find the given unit */
	id = (struct UnitId *) &libbase->vb_UnitList.lh_Head;
	while( id->ui_Node.ln_Succ && ( &id->ui_Unit != *ior->io_Unit ) ) {
		id = (struct UnitId *)id->ui_Node.ln_Succ;
	}
     
	if( id->ui_Node.ln_Succ ) {
    /* send CLOSE  message to the child process */
		struct MsgPort *OldReply;
    
		OldReply = ior->io_Message.mn_ReplyPort;
                DBUG(KPrintF("DevClose: OldReply\n"));
                   
		if( ( ior->io_Message.mn_ReplyPort = CreatePort( NULL, 0 ) ) ) {
			ior->io_Error = 0;

			ior->io_Command = FCMD_CLOSE;

			PutMsg( &id->ui_Unit.unit_MsgPort, &ior->io_Message );

			WaitPort( ior->io_Message.mn_ReplyPort );

			DeletePort( ior->io_Message.mn_ReplyPort );

		    //	  if( !id->ui_Unit.unit_OpenCnt )
                        {
				Unit_Free( libbase, id );
			}
	    }
		ior->io_Unit = (struct Unit **) - 1;

		ior->io_Message.mn_ReplyPort = OldReply;

#ifdef __MORPHOS__
		/* Free TLS data if allocated for this task. */

		if (tlsindex != TLS_INVALID_INDEX)
		{
			APTR value = TLSGetValue(tlsindex);
			if (value)
			{
				TLSSetValue(tlsindex, NULL); /* NOTE: NULL cannot fail */
				tlsdestructor(value);
			}
		}
#endif

		libbase->vb_Library.lib_OpenCnt--;
	}
        DBUG(KPrintF("DevClose: exit\n"));
#if !defined(__AROS__)        
        return 0;
#else
    return TRUE;
#endif
}

///
/// DevExpunge()

#if !defined(__AROS__)
BPTR DevExpunge( void)
{
	struct InstrumentBase *libbase = (APTR) REG_A6;
	BPTR ret = (BPTR) NULL;  /* default is no expunge */
#else
static int GM_UNIQUENAME(Expunge)
(
    LIBBASETYPEPTR libbase
)
{
#endif

	if( !libbase->vb_Library.lib_OpenCnt ) {   /* only expunge if opencnt is 0L */

#if !defined(__AROS__)
        ret = libbase->vb_SegList;
#endif

/* remove device from list */

        DBUG(KPrintF("dev expunge: list was empty\n"));

        Remove(&libbase->vb_Library.lib_Node);

/* free resources */

		DBUG(KPrintF("dev expunge: free resources\n"));

	   if( DOSBase) {
			CloseLibrary( (struct Library *) DOSBase );
		}
		
		// closing poseidon
			#ifdef __MORPHOS__
			if (PsdBase)
			{
				CloseLibrary(PsdBase);
				PsdBase = NULL;
			}
			
			#elif  __AROS__
			if (ps)
			{
				CloseLibrary(ps);
				ps = NULL;
			}
			#endif

			#ifdef __MORPHOS__
			/* Free the TLS index if one was in use. */

			if (tlsindex != TLS_INVALID_INDEX)
			{
				TLSFree(tlsindex);
			}
			#endif


/* free library memory */

        DBUG(KPrintF("dev expunge: remove done\n"));

		FreeMem((char *) libbase - libbase->vb_Library.lib_NegSize, (ULONG) (libbase->vb_Library.lib_NegSize + libbase->vb_Library.lib_PosSize ) );


        DBUG(KPrintF("dev expunge: FreeMem done\n"));

    } else {

		libbase->vb_Library.lib_Flags |= LIBF_DELEXP;

    }
#if !defined(__AROS__)    
    return(ret);
}
#else
    return TRUE;
}
#endif

///

#ifdef __AROS__

ADD2INITLIB(GM_UNIQUENAME(Init), 0)
ADD2EXPUNGELIB(GM_UNIQUENAME(Expunge), 0)
ADD2OPENDEV(GM_UNIQUENAME(Open), 0)
ADD2CLOSEDEV(GM_UNIQUENAME(Close), 0)
#endif

/// DevBeginIO()

#ifdef __MORPHOS__
void DevBeginIO(void)
{
    //struct InstrumentBase *libbase = (APTR) REG_A6;
    struct InstrumentIO *ior = (APTR) REG_A1;
    ior->io_Error = 0;

	if( (ior->io_Flags & IOF_QUICK) && Unit_QuickIO( ior ) )
    { /* quick mode? */
		ior->io_Message.mn_Node.ln_Type = NT_REPLYMSG;
	}
	else
    {
        ior->io_Message.mn_Node.ln_Type = NT_MESSAGE; /* add command to task port */
        PutMsg( &((*ior->io_Unit)->unit_MsgPort), &ior->io_Message );
    }
}

#elif __AROS__
AROS_LH1(void, beginio,
     AROS_LHA(struct IORequest *, ioreq, A1),
     struct InstrumentBase *, libbase, 5, Lexmark)
{
    AROS_LIBFUNC_INIT

    struct InstrumentIO *ior = (struct InstrumentIO*)ioreq;
    ior->io_Error = 0;

	if( (ior->io_Flags & IOF_QUICK) && Unit_QuickIO( ior ) )
    { /* quick mode? */
        ior->io_Message.mn_Node.ln_Type = NT_REPLYMSG;
	}
	else
    {
        ior->io_Message.mn_Node.ln_Type = NT_MESSAGE; /* add command to task port */
        PutMsg( &((*ior->io_Unit)->unit_MsgPort), &ior->io_Message );
    }
    AROS_LIBFUNC_EXIT
}

#endif

///
/// DevAbortIO()

#if !defined(__AROS__)
int DevAbortIO( void )
{
struct InstrumentBase *libbase = (APTR) REG_A6;
struct InstrumentIO *ior = (APTR) REG_A1;
 int result = 0;
#else
AROS_LH1(LONG, abortio,
    AROS_LHA(struct IORequest *,      ioreq, A1),
    struct InstrumentBase *, libbase, 6, Lexmark)
{
    AROS_LIBFUNC_INIT
    
    struct InstrumentIO *ior = (struct InstrumentIO*)ioreq;
#endif

	if( ( (ior->io_Flags & IOF_QUICK) || (ior->io_Message.mn_Node.ln_Type == NT_REPLYMSG ) ) ) {
		struct Node *np;
		BOOL Aborted = FALSE;

		Forbid();

		np = (*ior->io_Unit)->unit_MsgPort.mp_MsgList.lh_Head;
		while( np->ln_Succ && (np != &ior->io_Message.mn_Node ) ) {
			np = np->ln_Succ;
		}

		if( np == &ior->io_Message.mn_Node ) {
			Remove( np );
			ior->io_Error = IOERR_ABORTED;
			ReplyMsg( &ior->io_Message );
			Aborted = TRUE;
		}
		Permit();
	 
		if( !Aborted ) {
			struct UnitId *id;

			id = (struct UnitId *) &libbase->vb_UnitList.lh_Head;
			while( id->ui_Node.ln_Succ && ( &id->ui_Unit != *ior->io_Unit ) ) {
				id = (struct UnitId *)id->ui_Node.ln_Succ;
			}

			/*if( id->ui_Node.ln_Succ ) {
			// send ABORT signal to the child process
				Signal( &id->ui_Prc->pr_Task, SIGBREAKF_CTRL_C );
			}*/
		}
	}
#if !defined(__AROS__)
    return 0;
}
#else
	return 0;
    
    AROS_LIBFUNC_EXIT
}
#endif

///

#ifdef __MORPHOS__

// have to fix this
void exit(int x)
{
	kprintf("FATAL SANE ERROR: exit(%d) called - task %p hung\n", x, FindTask(NULL));
	for (;;)
		Wait(SIGF_SINGLE);
}

/*
 * TLS implementation to hold errno and OSERR
 */

struct __tls_variables
{
	int tls_errno;
	int tls_OSERR;
};

static struct __tls_variables *__gettls(void)
{
	static struct __tls_variables statictls;
	struct __tls_variables *tls = NULL;

	if (tlsindex != TLS_INVALID_INDEX)
	{
		tls = TLSGetValue(tlsindex);
		if (!tls)
		{
			tls = AllocMem(sizeof(*tls), MEMF_CLEAR);
			if (tls)
			{
				if (!TLSSetValue(tlsindex, tls))
				{
					FreeMem(tls, sizeof(*tls));
					tls = NULL;
				}
			}
		}
	}

	/* Fall back to static data in case something fails */
	if (!tls)
		tls = &statictls;

	return tls;
}

int *__errno_location(void)
{
	struct __tls_variables *tls = __gettls();
	return &tls->tls_errno;
}

int *__OSERR_location(void)
{
	struct __tls_variables *tls = __gettls();
	return &tls->tls_OSERR;
}

static VOID tlsdestructor(APTR value)
{
	struct __tls_variables *tls = value;
	FreeMem(tls, sizeof(*tls));
}

#endif

